import copy
from io import BytesIO

import pytest

from fittie.fitfile.data_message import (
    decode_data_message,
    add_subfields_to_fields,
    apply_scale_and_offset,
)
from fittie.fitfile.profile.util import get_message_profile
from fittie.fitfile.records import RecordHeader


def test_decode_data_message(record_1_definition_message):
    # Expect record 2
    header = RecordHeader(
        is_definition_message=False,
        is_developer_data=False,
        is_compressed_timestamp_message=False,
        local_message_type=0,
    )
    data = b"\x04\x0f\x00\x16\x00\xd2\x04\x00\x00(\xc6\n%"
    data_message = decode_data_message(
        header, record_1_definition_message, developer_data={}, data=BytesIO(data)
    )
    assert data_message.fields == {
        "type": 4,
        "manufacturer": 15,
        "product": 22,
        "serial_number": 1234,
        "time_created": 621463080,
        "garmin_product": 22,  # Subfield
    }


def test_decode_data_message_with_developer_fields(record_5_definition_message):
    # Expect record 6
    header = RecordHeader(
        is_definition_message=False,
        is_developer_data=False,
        is_compressed_timestamp_message=False,
        local_message_type=0,
    )
    data = (
        b"\x00\x00\x01"
        b"doughnuts earned\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"doughnuts\x00\x00\x00\x00\x00\x00\x00"
    )
    data_message = decode_data_message(
        header, record_5_definition_message, developer_data={}, data=BytesIO(data)
    )
    assert data_message.fields == {
        "developer_data_index": 0,
        "fit_base_type_id": 1,
        "field_definition_number": 0,
        "field_name": "doughnuts earned",
        "units": "doughnuts",
    }


def test_add_subfields_to_fields():
    fields = {
        "type": 9,
        "manufacturer": 15,
        "product": 22,
        "serial_number": 1234,
        "time_created": 621463080,
    }
    fields_raw = copy.deepcopy(fields)
    field_profile = get_message_profile(0).fields[2]
    fields_with_components = []
    subfield_names = add_subfields_to_fields(
        fields, fields_raw, field_profile, fields_with_components
    )
    assert subfield_names == ["garmin_product"]
    assert "garmin_product" in fields
    assert fields["product"] == fields["garmin_product"]


@pytest.mark.parametrize(
    "value,scale,offset,expected",
    [
        (123, 10, 1, 11.3),
        ([1, 2], 10, 0, [0.1, 0.2]),
        ([1, 2], 10, 1, [-0.9, -0.8]),
        ([1, 2], [10, 100], 0, [0.1, 0.02]),
        (None, 10, 10, None),
        ([None, None], 10, 10, [None, None]),
    ],
)
def test_apply_scale_and_offset(value, scale, offset, expected):
    assert apply_scale_and_offset(value, scale, offset) == expected



from dataclasses import fields
from functools import lru_cache

from fittie.fitfile.profile.message_profile import (
    MessageProfile,
    FieldProfile,
    SubField,
)
from fittie.fitfile.profile.messages import MESSAGES


@lru_cache(maxsize=None)
def get_message_profile(number: int) -> MessageProfile:
    """
    A cached helper method to retrieve data from messages.py as a MessageProfile by
    providing the message profile number.


    Notes:
        Replace lru_cache with cache when minimum Python version is 3.9
    """
    if number not in MESSAGES:
        raise ValueError(f'unknown message number "{number}"')

    return dict_to_message_profile(MESSAGES[number])


def dict_to_message_profile(source: dict) -> MessageProfile:
    """
    Converts a nested dict to a MessageProfile, with FieldProfiles
    """
    raw = {}
    field_types = {f.name: f.type for f in fields(MessageProfile)}

    for key, value in source.items():
        if field_types[key] != "dict[int, FieldProfile]":
            raw[key] = value
            continue

        raw_fields = {}
        for field_key, field_value in value.items():
            raw_fields[field_key] = dict_to_field_profile(field_value)

        raw[key] = raw_fields

    return MessageProfile(**raw)


def dict_to_field_profile(source: dict) -> FieldProfile:
    """
    Converts a nested dict to a FieldProfile, with SubFields
    """
    raw = {}
    field_types = {f.name: f.type for f in fields(FieldProfile)}

    for key, value in source.items():
        if field_types[key] != "Optional[list[SubField]]":
            raw[key] = value
            continue

        raw_subfields = []
        for subfield in value:
            raw_subfields.append(SubField(**subfield))

        raw[key] = raw_subfields

    return FieldProfile(**raw)



from __future__ import annotations  # Added for type hints
import logging
from copy import deepcopy
from typing import Any, Optional, TYPE_CHECKING

from fittie.fitfile.profile import FieldProfile
from fittie.fitfile.utils.datastream import Streamable
from fittie.fitfile.utils.exceptions import DecodeException
from fittie.fitfile.definition_message import DefinitionMessage
from fittie.fitfile.field_definitions import read_field, read_developer_field
from fittie.fitfile.field_description import FieldDescription
from fittie.fitfile.profile.util import get_message_profile

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from fittie.fitfile.records import RecordHeader


class DataMessage:
    """
    Contains a local message type and populated data fields as described by the
    preceding definition message.

    There are two types of data message:
        Normal Data Message
        Compressed Timestamp Data Message

    Related DefinitionMessages and DataMessages share a local message type
    """

    header: "RecordHeader"
    fields: dict[str, Optional[Any]]

    def __init__(self, header: "RecordHeader", fields: dict[str, Optional[Any]]):
        self.header = header
        self.fields = fields

    def get_field(self, field_name: str) -> Optional[Any]:
        """Retrieve a field by key from fields"""
        return self.fields.get(field_name, None)

    @property
    def local_message_type(self) -> int:
        """Retrieves the local message type from the message header"""
        return self.header.local_message_type

    def __str__(self) -> str:
        return f"DataMessage:{self.fields}".replace("self.", " ")

    def __repr__(self) -> str:
        return str(self)


def add_subfields_to_fields(
    fields: dict[str, Any],
    fields_raw: dict[str, Any],
    field_profile: FieldProfile,
    fields_with_components: list[str],
) -> list[str]:
    """
    Adds the field value of a DataMessage field that matches the provided field profile
    as an extra field to the DataMessage fields.

    For each subfield, it checks if the field name exists in the current fields dict,
    if so it checks if the field value equals the reference value number.

    If it finds a match, it adds the original field value to the fields dict, with the
    subfield name as key. The original value is retrieved from
    fields_raw, these values are not affected by scale / offset.

    For example {"product": 22} as fields, becomes {"product": 22, "garmin_product": 22}
    """

    subfield_names = []

    for subfield in field_profile.subfields:
        for reference in subfield.refs:
            if not (field_value := fields.get(reference["field_name"])):
                continue

            if reference["value_number"] == field_value:
                subfield_names.append(subfield.field_name)
                field_data = deepcopy(fields_raw[field_profile.field_name])

                if subfield.scale is not None or subfield.offset is not None:
                    field_data = apply_scale_and_offset(
                        field_data,
                        subfield.scale,
                        subfield.offset
                    )

                fields[subfield.field_name] = field_data

                if subfield.has_components:
                    fields_with_components.append(subfield.field_name)

    return subfield_names


def apply_scale_and_offset(
    field_data: Any, scale: Optional[int], offset: Optional[int]
) -> Any:
    """
    Applies scale and offset to the provided value

    Value is divided by scale, default scale is 1
    Value is subtracted by ofset, default offset is 0

    Value can be a single value, or list of value
    Scale can be a single value, or list of value
    Offset is single value
    """
    if field_data is None:
        return field_data

    scale = scale or 1
    offset = offset or 0

    if not isinstance(field_data, list):
        return field_data / scale - offset

    if isinstance(scale, list):
        if len(scale) != len(field_data):
            return field_data

        return [apply_scale_and_offset(d, s, offset) for d, s in zip(field_data, scale)]

    return [apply_scale_and_offset(data, scale, offset) for data in field_data]


def decode_data_message(
    header: "RecordHeader",
    message_definition: DefinitionMessage,
    developer_data: dict[int, dict[str, dict[int, FieldDescription]]],
    data: Streamable,
) -> DataMessage:
    message_profile = get_message_profile(message_definition.global_message_type)

    fields: dict[str, Any] = {}
    # Field values without scale and offset applied,
    # used for subfields
    fields_raw: dict[str, Any] = {}
    fields_with_subfields: dict[str, FieldProfile] = {}
    fields_with_components: list[str] = []
    subfield_names = []

    for field in message_definition.field_definitions:
        field_profile = message_profile.fields[field.number]
        field_data = read_field(
            field_definition=field,
            endianness=message_definition.endianness,
            data=data,
        )
        fields_raw[field_profile.field_name] = field_data

        if field_profile and (
            field_profile.scale is not None or field_profile.offset is not None
        ):
            field_data = apply_scale_and_offset(
                field_data, scale=field_profile.scale, offset=field_profile.offset
            )

        fields[field_profile.field_name] = field_data

        if field_profile.has_subfields:
            fields_with_subfields[field_profile.field_name] = field_profile
        if field_profile.has_components:
            fields_with_components.append(field_profile.field_name)

    # TODO: components, accumulate etc (from field_profile?)
    if fields_with_subfields:
        for field, field_profile in fields_with_subfields.items():
            subfield_names += add_subfields_to_fields(
                fields, fields_raw, field_profile, fields_with_components
            )
    if fields_with_components:
        logger.debug(f"components not implemented yet, {fields_with_components=}")

    # if field_profile.accumulate:
    #     ...

    if message_definition.developer_field_definitions and not developer_data:
        raise DecodeException(
            detail="definition message contains developer fields, "
            "but no field descriptions are provided",
            position=data.tell(),
        )

    for field in message_definition.developer_field_definitions:
        try:
            field_description = developer_data[field.data_index]["fields"][field.number]
        except KeyError:
            raise DecodeException(
                detail=f"no field description found for field {field}",
                position=data.tell(),
            )

        developer_field_definition = message_definition.get_developer_field_definition(
            data_index=field_description.developer_data_index,
            number=field_description.field_definition_number,
        )

        if not developer_field_definition:
            raise DecodeException(
                detail=f"no developer field definition found for field {field}",
                position=data.tell(),
            )

        field_data = read_developer_field(
            field_description=field_description,
            field_definition=developer_field_definition,
            endianness=message_definition.endianness,
            data=data,
        )

        fields[field_description.field_name] = field_data

    return DataMessage(header=header, fields=fields)



import pytest

from fittie.utils.gear_change_data import get_gear_change_data


@pytest.mark.parametrize(
    "data,expected",
    [
        (
            654380552,
            {
                "front_gear": 39,
                "front_gear_number": 1,
                "rear_gear": 14,
                "rear_gear_number": 8,
            },
        ),
        (
            654380041,
            {
                "front_gear": 39,
                "front_gear_number": 1,
                "rear_gear": 12,
                "rear_gear_number": 9,
            },
        ),
        (
            872549385,
            {
                "front_gear": 52,
                "front_gear_number": 2,
                "rear_gear": 12,
                "rear_gear_number": 9,
            },
        ),
        (
            0,
            {
                "front_gear": 0,
                "front_gear_number": 0,
                "rear_gear": 0,
                "rear_gear_number": 0,
            },
        ),
    ],
)
def test_get_gear_change_data(data, expected):
    assert get_gear_change_data(data) == expected



from datetime import timezone

import pytest

from fittie import decode


@pytest.mark.parametrize("load_fit_file", ["fittie_minimal_file.fit"], indirect=True)
def test_minimal_file(load_fit_file):
    fitfile = decode(load_fit_file)

    header = fitfile.header
    assert header.length == 14
    assert header.protocol_version == 32
    assert header.profile_version == 2033
    assert header.data_size == 111

    developer_data = fitfile.developer_data.get(1)
    assert developer_data is not None
    cleaned_application_id = [num for num in developer_data["application_id"] if num]
    assert bytes(cleaned_application_id) == b"com.marcelblijleven.fittie"
    assert developer_data["application_version"] == 1337

    file_id = fitfile.file_id
    assert file_id["type"] == "activity"
    assert file_id["manufacturer"] == "development"
    assert file_id["product_name"] == "fittie-test"
    assert file_id["time_created"].tzinfo == timezone.utc
    ...


@pytest.mark.parametrize("load_fit_file", ["fittie_gearshifts.fit"], indirect=True)
def test_gearshifts(load_fit_file):
    fitfile = decode(load_fit_file)
    event_messages = fitfile.get_messages_by_type("event")
    assert len(event_messages) == 3

    # NOTE: this isn't exactly as expected, the field gear_change_data has
    # subfields and components, so it should be expanded into 4 fields.
    # Have not figured out a fix yet, could be an error in the Profile.xlsx
    # It expects a field name called 'event', but it is filled in as 'data'.
    assert event_messages[0].fields["data"] == 654380552
    assert event_messages[1].fields["data"] == 654380041
    assert event_messages[2].fields["data"] == 872549385


@pytest.mark.parametrize(
    "load_fit_file", ["fittie_developer_fields.fit"], indirect=True
)
def test_developer_fields(load_fit_file):
    fitfile = decode(load_fit_file)
    assert len(fitfile.developer_data.keys()) == 2

    fields = fitfile.developer_data[0]["fields"]
    assert (banana_field := fields[0]).field_name == "bananas_traversed"
    assert banana_field.units == "bananas"


@pytest.mark.parametrize("load_fit_file", ["fittie_settings_file.fit"], indirect=True)
def test_gearshifts(load_fit_file):
    fitfile = decode(load_fit_file)
    assert fitfile.file_type == "settings"
    assert (user_profile_messages := fitfile.data_messages.get("user_profile")) is not None
    assert user_profile_messages[0].fields["gender"] == 1
    assert user_profile_messages[0].fields["age"] == 33
    assert user_profile_messages[0].fields["weight"] == 80.1
    assert user_profile_messages[0].fields["friendly_name"] == "Fittie McFitface"


@pytest.mark.parametrize("load_fit_file", ["fittie_monitoring_file.fit"], indirect=True)
def test_monitoring_file(load_fit_file):
    """NOTE: This file has accumulated fields"""
    fitfile = decode(load_fit_file)
    assert fitfile.file_type == "monitoring_b"

    monitoring_messages = fitfile.get_messages_by_type("monitoring")

    assert len(monitoring_messages) == 24  # 24 hours

    for message in monitoring_messages:
        # Test if subfield has "scale" applied correctly
        assert message.fields["cycles"] * 2 == message.fields["steps"]



import pytest

from datetime import datetime, timezone

from fittie.utils.enrich_data import enrich_data


@pytest.mark.parametrize("data,expected", [
    ({"timestamp": 1045945633}, {"timestamp": datetime(2023, 2, 21, 20, 27, 13, tzinfo=timezone.utc)}),
    ({"heart_rate": 140, "manufacturer": 255}, {"heart_rate": 140, "manufacturer": "development"}),
    ({"language": 8}, {"language": "dutch"}),
    ({"i_do_not_exist": 1337}, {"i_do_not_exist": 1337}),
    ({"time_zone": 40}, {"time_zone": "amsterdam"}),
])
def test_enrich_data(data, expected):
    enrich_data(data)
    assert data == expected



from __future__ import annotations  # added for type hints

from dataclasses import dataclass
from typing import Union, Optional


@dataclass
class SubField:
    field_name: str
    field_type: str
    refs: list[dict[str, Union[int, str]]]
    array: Optional[Union[bool, str, int]] = None
    components: Optional[Union[str, list[str]]] = None
    scale: Optional[Union[int, list[int]]] = None
    offset: Optional[int] = None
    units: Optional[str] = None
    bits: Optional[Union[int, str]] = None
    accumulate: Optional[Union[int, list[int]]] = None
    comment: Optional[str] = None
    ref_field_name: Optional[Union[str, list[str]]] = None
    ref_field_value: Optional[Union[str, list[str]]] = None

    @property
    def is_array(self) -> bool:
        """Check whether the field is an array"""
        return self.array is not None

    @property
    def has_components(self) -> bool:
        """Check whether the field has components"""
        return self.components is not None


@dataclass
class FieldProfile:
    field_name: str
    field_type: str
    array: Optional[Union[bool, str, int]] = None
    components: Optional[Union[str, list[str]]] = None
    scale: Optional[Union[int, list[int]]] = None
    offset: Optional[int] = None
    units: Optional[str] = None
    bits: Optional[Union[int, str]] = None
    accumulate: Optional[Union[int, list[int]]] = None
    subfields: Optional[list[SubField]] = None
    comment: Optional[str] = None

    @property
    def is_array(self) -> bool:
        """Check whether the field is an array"""
        return self.array is not None

    @property
    def has_components(self) -> bool:
        """Check whether the field has components"""
        return self.components is not None

    @property
    def has_subfields(self) -> bool:
        """Check whether the field has subfields"""
        return self.subfields and len(self.subfields) > 0


@dataclass
class MessageProfile:
    name: str
    fields: dict[int, FieldProfile]
    group: str
